MODULE tree

  TYPE vector
    DOUBLE PRECISION         :: vec(2)
  END TYPE vector

  TYPE matrix
    DOUBLE PRECISION         :: mat(2,2)
  END TYPE matrix

! Here we define the data type NODE which will contain
  TYPE node
    INTEGER                  :: level         ! level in tree: top=m, bottom=1
    INTEGER                  :: lags(26)      ! lags for node are stored here
    TYPE (matrix)            :: A(26)         ! forward coefficients for node
    TYPE (matrix)            :: B(26)         ! backward coefficients for node
    TYPE (matrix)            :: vf, vb        ! forward and backward MSEs
    TYPE (vector)            :: eps(1:10100)  ! the epsilons for the node
    TYPE (vector)            :: eta(-99:10000)! the etas for the node
    TYPE (node), POINTER     :: reg, star     ! pointers to the regular and
  END TYPE node                               ! starred subnodes one level down

! These will contain the end results
    TYPE (matrix)            :: topA(1:26), topvf
! Other globals
  INTEGER                    :: n, m, method
  INTEGER, ALLOCATABLE       :: orig_lags(:)
  TYPE (vector), ALLOCATABLE :: x(:)
  CHARACTER                  :: stamp*4


CONTAINS

SUBROUTINE make_tree(toplags, truevf)
  INTEGER                    :: toplags(m)
  TYPE (node), POINTER       :: top_node
  DOUBLE PRECISION           :: truevf(3)
  EXTERNAL DEVLRG

  NULLIFY (top_node)                  !associates top_node so we can use it
  ALLOCATE(orig_lags(m))
  orig_lags=toplags

! now build the tree of node lags
  CALL build_node_tree(m,toplags,top_node)  

! now fill the tree, i.e. get coeffts and MSE's of each node
  CALL fill_tree(top_node)

! node tree built, so print it
  CALL print_node_tree(top_node) 

! undo node tree so that we don't run out of memory for next runs 
! CALL undo_node_tree(top_node)

! Causal check - program will terminate if noncausal solution.
  CALL Causal_Check

! Likelihood calculations (exact or approx)
!  CALL likelihood("YW WN")
  CALL approx_likelihoods(truevf)

  RETURN
END SUBROUTINE make_tree

!***********************************************************************

RECURSIVE SUBROUTINE build_node_tree(lev,this_lags,this_node)
  INTEGER                 :: i, lev, this_lags(26)
  TYPE (node), POINTER    :: this_node  

! This routine will create a tree of nodes needed to subset Burg model. The
! level and lags are assigned to each node. Also the MSE of each node is
! initialized to be zero so that later we'll be able to check which nodes
! have not yet been filled.

! first time thru' with a fresh node; point to it & make its J and J*
! pointers point nowhere
  IF (.NOT. ASSOCIATED(this_node)) THEN
    ALLOCATE (this_node)
    this_node%level=lev  
    this_node%lags(1:lev)=this_lags(1:lev)
!   the check for an unfilled node will be the (1,1) entry of vf=0
    this_node%vf%mat(1,1)=0           
    NULLIFY (this_node%reg) 
    NULLIFY (this_node%star)  
  END IF

! recursive call to routine with J lags; only if level>1
  lev=this_node%level
  IF (lev>1) THEN 
    this_lags(1:lev-1)=this_node%lags(1:lev-1)                   
    CALL build_node_tree(lev-1,this_lags,this_node%reg)
  END IF

! recursive call to routine with J* lags; only if level>1
  lev=this_node%level
  IF (lev>1) THEN  
    this_lags(1:lev-1)=(/(this_node%lags(lev)              &
                                 -this_node%lags(lev-i), i=1,lev-1)/)   
    CALL build_node_tree(lev-1,this_lags,this_node%star)  
  END IF

  RETURN
END SUBROUTINE build_node_tree
 
!***********************************************************************

RECURSIVE SUBROUTINE fill_tree(this_node)
  INTEGER               :: i
  TYPE (node), POINTER  :: this_node  

! Here we fill in the coeffts and MSE for each node

  IF (this_node%level>1) THEN
    CALL fill_tree(this_node%reg)
!   if reg node has not been filled, then fill it!
    IF (this_node%reg%vf%mat(1,1)==0) THEN 
      CALL fill_node(this_node%reg)
    END IF

    CALL fill_tree(this_node%star)

!   if star node has not been filled, then fill it!
    IF (this_node%star%vf%mat(1,1)==0) THEN 
      CALL fill_node(this_node%star)
    END IF
  END IF

! now that whole tree is filled, we can fill top node
  IF (this_node%level==m) THEN
    CALL fill_node(this_node)
  END IF

  RETURN
END SUBROUTINE fill_tree

!***********************************************************************

SUBROUTINE fill_node(this_node)
  INTEGER               :: i, j, t, km, lev, range(2)
  TYPE (vector), ALLOCATABLE       :: eps_J(:),eps_Js(:),eta_J(:),eta_Js(:)
  DOUBLE PRECISION, DIMENSION(2,2) :: vf_J, vf_Js, A_K, B_Ks, sen, see, Id
  DOUBLE PRECISION, DIMENSION(2,2) :: vfinv, vfsinv, B, A_Ks, B_K,vb_J,vb_Js
  DOUBLE PRECISION, DIMENSION(2,2) :: vbinv, vbsinv, snn, R, see2, snn2
  DOUBLE PRECISION, DIMENSION(2,2) :: U2, V2, AA, CC, tV
  DOUBLE PRECISION, DIMENSION(4,4) :: term1, term1_left, term1_right, A
  TYPE (matrix)         :: A_J(26), B_Js(26), A_Js(26), B_J(26) 
  TYPE (node), POINTER  :: this_node  

! This is the routine where the real work of building the coeffts, MSEs,
! epsilons and etas is done

! initialize the precursors before applying algo to this node
  lev=this_node%level
  km=this_node%lags(lev)
  ALLOCATE(eps_J(1:n+km),eps_Js(1:n+km),eta_J(1-km:n),eta_Js(1-km:n))
  IF (lev==1) THEN 
    DO t=n+1,n+km
      eps_J(t)%vec =0; eps_Js(t)%vec=0 
    END DO
    DO t=1-km,0
      eta_J(t)%vec =0; eta_Js(t)%vec=0  
    END DO
    eps_J (1:n)=x(1:n)
    eps_Js(1:n)=x(1:n)
    eta_J (1:n)=x(1:n)
    eta_Js(1:n)=x(1:n)
    DO i=1,2
      DO j=1,2
         vf_J(i,j)=DOT_PRODUCT(x%vec(i),x%vec(j))/n
      END DO
    END DO
    vf_Js  =vf_J
    vb_Js  =vf_J
    vb_J  =vf_J
  ELSE  !we're at a higher level, so use reg and star node info 
    A_J (1:lev-1)  =this_node%reg%A(1:lev-1)
    A_Js(1:lev-1)  =this_node%star%A(1:lev-1)
    B_Js(1:lev-1)  =this_node%star%B(1:lev-1)
    B_J (1:lev-1)  =this_node%reg%B(1:lev-1)
    eps_J (1:n+km) =this_node%reg%eps(1:n+km)
    eps_Js(1:n+km) =this_node%star%eps(1:n+km)    
    eta_J (1-km:n) =this_node%reg%eta(1-km:n)
    eta_Js(1-km:n) =this_node%star%eta(1-km:n)
    vf_J  =this_node%reg%vf%mat
    vf_Js =this_node%star%vf%mat
    vb_Js =this_node%star%vb%mat
    vb_J  =this_node%reg%vb%mat
  END IF

! Initial conditions set, now apply algo to this node
! First build A_K(km) -------------------------------
  range=(/1,n+km/)
  IF (method>1) range=(/1+km,n/)
  DO i=1,2
    see(i,:)=(/0.0D+00,0.0D+00/)
    sen(i,:)=(/0.0D+00,0.0D+00/)
    snn(i,:)=(/0.0D+00,0.0D+00/)
  END DO
  DO t=range(1),range(2)
    DO i=1,2
      DO j=1,2
        see(i,j)=see(i,j)+eps_J(t)%vec(i)*eps_J(t)%vec(j)
        sen(i,j)=sen(i,j)+eps_J(t)%vec(i)*eta_Js(t-km)%vec(j)
        snn(i,j)=snn(i,j)+eta_Js(t-km)%vec(i)*eta_Js(t-km)%vec(j)
      END DO
    END DO
  END DO
! form I
  Id(1,:)=(/1.0D+00,0.0D+00/)
  Id(2,:)=(/0.0D+00,1.0D+00/)
! get the inverse of vf_J, put into vfinv
  CALL DLINRG(2,vf_J,2,vfinv,2)
  SELECT CASE (method)
    CASE (1)
!     get the inverse of vb_Js, put into vbsinv
      CALL DLINRG(2,vb_Js,2,vbsinv,2)  
!     finally, get A_K(km) & B_Ks(km)
      A_K =MATMUL(sen, vbsinv)/FLOAT(n)
      B_Ks=MATMUL(TRANSPOSE(sen), vfinv)/FLOAT(n)
    CASE (2)
      B=sen+MATMUL(MATMUL(vfinv,sen),vb_Js)
      CALL KRON(2,snn,Id,term1_left)
      CALL KRON(2,MATMUL(vb_Js,vb_Js),MATMUL(MATMUL(vfinv,see),vfinv), &
       &      term1_right)
      term1=term1_left+term1_right
      CALL DLINRG(4,term1,4,A,4)
!     finally, build elements of A_K piecemeal
      DO i=1,2
        A_K(i,1)=A(i,1)*B(1,1)+A(i,2)*B(2,1)+A(i,3)*B(1,2)+A(i,4)*B(2,2)
      END DO
      DO i=3,4
        A_K(i-2,2)=A(i,1)*B(1,1)+A(i,2)*B(2,1)+A(i,3)*B(1,2)+A(i,4)*B(2,2)
      END DO
!     Finished building A_K -------------- now get B_Ks from it
      B_Ks=MATMUL(MATMUL(vb_Js,TRANSPOSE(A_K)),vfinv)
    CASE(3)
      CALL Matrix_Power(vf_J, 5.0D-1, U2)
      CALL Matrix_Power(vb_Js, -5.0D-1, V2)
      CALL Matrix_Power(see, -5.0D-1, see2)
      CALL Matrix_Power(snn, -5.0D-1, snn2)
      R=MATMUL(MATMUL(see2,sen),snn2)
      A_K=MATMUL(MATMUL(U2,R),V2)
      B_Ks=MATMUL(MATMUL(vb_Js,TRANSPOSE(A_K)),vfinv)
    CASE(4)
      CALL DLINRG(2,vb_Js,2,V2,2) ! V2=(V_J*)^-1
      U2=vfinv                    ! U2=(U_J)^-1
      B=MATMUL(snn,V2)
      AA=MATMUL(see,U2)
      CALL KRON(2,Id,AA,term1_left)
      CALL KRON(2,B,Id,term1_right)
      term1=term1_left+term1_right
      CALL DLINRG(4,term1,4,A,4)
      CC=2.0*sen
!     Now vec(R) = A . vec(CC)
      B=CC
!     finally, build elements of R piecemeal
      DO i=1,2
        R(i,1)=A(i,1)*B(1,1)+A(i,2)*B(2,1)+A(i,3)*B(1,2)+A(i,4)*B(2,2)
      END DO
      DO i=3,4
        R(i-2,2)=A(i,1)*B(1,1)+A(i,2)*B(2,1)+A(i,3)*B(1,2)+A(i,4)*B(2,2)
      END DO
!     Finished building R --------------  now get A_K & B_Ks from it
      A_K=MATMUL(R,V2)
      B_Ks=MATMUL(MATMUL(vb_Js,TRANSPOSE(A_K)),vfinv)
  END SELECT

! now get A_Ks, just swap star and nostar
  DO i=1,2
    see(i,:)=(/0.0D+00,0.0D+00/)
    sen(i,:)=(/0.0D+00,0.0D+00/)
    snn(i,:)=(/0.0D+00,0.0D+00/)
  END DO
  DO t=range(1),range(2)
    DO i=1,2
      DO j=1,2
        see(i,j)=see(i,j)+eps_Js(t)%vec(i)*eps_Js(t)%vec(j)
        sen(i,j)=sen(i,j)+eps_Js(t)%vec(i)*eta_J(t-km)%vec(j)
        snn(i,j)=snn(i,j)+eta_J(t-km)%vec(i)*eta_J(t-km)%vec(j)
      END DO
    END DO
  END DO    
! get the inverse of vf_Js, put into vfsinv
  CALL DLINRG(2,vf_Js,2,vfsinv,2)
  SELECT CASE (method)
    CASE (1)
!     get the inverse of vb_J, put into vbinv
      CALL DLINRG(2,vb_J,2,vbinv,2) 
!     finally, get A_Ks(km) & B_K(km)
      A_Ks=MATMUL(sen, vbinv)/FLOAT(n)
      B_K =MATMUL(TRANSPOSE(sen), vfsinv)/FLOAT(n)
    CASE (2)
      B=sen+MATMUL(MATMUL(vfsinv,sen),vb_J)
      CALL KRON(2,snn,Id,term1_left)
      CALL KRON(2,MATMUL(vb_J,vb_J),MATMUL(MATMUL(vfsinv,see),vfsinv), &
        &     term1_right)
      term1=term1_left+term1_right
      CALL DLINRG(4,term1,4,A,4)
!     finally, build elements of A_Ks piecemeal
      DO i=1,2
        A_Ks(i,1)=A(i,1)*B(1,1)+A(i,2)*B(2,1)+A(i,3)*B(1,2)+A(i,4)*B(2,2)
      END DO
      DO i=3,4
        A_Ks(i-2,2)=A(i,1)*B(1,1)+A(i,2)*B(2,1)+A(i,3)*B(1,2)+A(i,4)*B(2,2)
      END DO
!     Finished building A_Ks ---------------
!     now get B_K from it
      B_K=MATMUL(MATMUL(vb_J,TRANSPOSE(A_Ks)),vfsinv)
    CASE(3)
      CALL Matrix_Power(vf_Js, 5.0D-1, U2)
      CALL Matrix_Power(vb_J, -5.0D-1, V2)
      CALL Matrix_Power(see, -5.0D-1, see2)
      CALL Matrix_Power(snn, -5.0D-1, snn2)
      R=MATMUL(MATMUL(see2,sen),snn2)
      A_Ks=MATMUL(MATMUL(U2,R),V2)
      B_K=MATMUL(MATMUL(vb_J,TRANSPOSE(A_Ks)),vfsinv)
    CASE(4)
      CALL DLINRG(2,vb_J,2,V2,2) ! V2=(V_J)^-1
      U2=vfsinv                  ! U2=(U_J*)^-1
      B=MATMUL(snn,V2)
      AA=MATMUL(see,U2)
      CALL KRON(2,Id,AA,term1_left)
      CALL KRON(2,B,Id,term1_right)
      term1=term1_left+term1_right
      CALL DLINRG(4,term1,4,A,4)
      CC=2.0*sen
!     Now vec(R) = A . vec(CC)
      B=CC
!     finally, build elements of R piecemeal
      DO i=1,2
        R(i,1)=A(i,1)*B(1,1)+A(i,2)*B(2,1)+A(i,3)*B(1,2)+A(i,4)*B(2,2)
      END DO
      DO i=3,4
        R(i-2,2)=A(i,1)*B(1,1)+A(i,2)*B(2,1)+A(i,3)*B(1,2)+A(i,4)*B(2,2)
      END DO
!     Finished building R --------------  now get A_Ks & B_K from it
      A_Ks=MATMUL(R,V2)
      B_K=MATMUL(MATMUL(vb_J,TRANSPOSE(A_Ks)),vfsinv)
  END SELECT


! assign A_K and B_K to correct node place
  this_node%A(lev)%mat=A_K
  this_node%B(lev)%mat=B_K
! Now do eqtns 1.16 and 1.18
  IF (lev > 1) THEN
    DO i=1,lev-1
      this_node%A(i)%mat=A_J(i)%mat-MATMUL(A_K,B_Js(lev-i)%mat)
      this_node%B(i)%mat=B_J(i)%mat-MATMUL(B_K,A_Js(lev-i)%mat)
    END DO
  END IF

! set MSE's, eqtns 1.19 and 1.20
  this_node%vf%mat=MATMUL((Id-MATMUL(A_K,B_Ks)),vf_J)
  this_node%vb%mat=MATMUL((Id-MATMUL(B_K,A_Ks)),vb_J)

! Set epsilons and etas
  DO i=range(1),range(2)
    this_node%eps(i)%vec   =eps_J(i)%vec-MATMUL(A_K,eta_Js(i-km)%vec)
    this_node%eta(i-km)%vec=eta_J(i-km)%vec-MATMUL(B_K,eps_Js(i)%vec)
  END DO

  DEALLOCATE(eps_J, eps_Js ,eta_J ,eta_Js)

  RETURN
END SUBROUTINE fill_node

!***********************************************************************

RECURSIVE SUBROUTINE print_node_tree(this_node)
  INTEGER               :: i, km, t, j
  TYPE (node), POINTER  :: this_node  

! Here we print the info in the top node - can also make it print all nodes
! by removing the inmost IF THEN loop

  IF (ASSOCIATED(this_node)) THEN
    CALL print_node_tree(this_node%reg)
    IF (this_node%level==m) THEN
      PRINT*,"***************************************************"
      PRINT*,"Estimated subset ",stamp," coefficient matrices:"
      DO i=1,this_node%level
        PRINT*,"Phi(", this_node%lags(i),"):"
        CALL DWRRRL('',2,2,this_node%A(i)%mat,2,0, &
                   '(F20.6)','NONE','NONE')
      END DO
      PRINT*," "
      PRINT*,"***************************************************"
      PRINT*,"Estimated ",stamp," (forward) WN covariance matrix:"
      CALL DWRRRL('',2,2,this_node%vf%mat,2,0,'(F20.6)','NONE','NONE')
      PRINT*," "

!     Store results for likelihood calcs
      topA(1:m)=this_node%A(1:m)
      topvf=this_node%vf
    END IF
    CALL print_node_tree(this_node%star)
    IF (this_node%level<m) DEALLOCATE (this_node)
  END IF

  RETURN
END SUBROUTINE print_node_tree

!***********************************************************************

SUBROUTINE Matrix_Power(A, pow, B)
! Raises (2 by 2) matrix A to a real power pow, result into matrix B.

  INTEGER          :: ipath, irank
  DOUBLE PRECISION :: pow, tol, A(2,2), B(2,2), D(2,2), U(2,2), V(2,2), eig(2)

  ipath=11
  tol=1.0D-10

  CALL DLSVRR(2, 2, A, 2, ipath, tol, irank, eig, U, 2, V, 2)
  D=0.0
  D(1,1)=eig(1)**ABS(pow); D(2,2)=eig(2)**ABS(pow)

  B=MATMUL(MATMUL(U,D),TRANSPOSE(V))
  IF (pow<0) CALL DLINRG(2,B,2,B,2)

  RETURN
END SUBROUTINE Matrix_Power
 
!***********************************************************************

SUBROUTINE KRON (d,A,B,C)
! computes the Kronecker product of square matrices A and B (dim=d), 
! puts into C (dim=d^2)
  INTEGER             :: d, i, j, k, l, row, col
  DOUBLE PRECISION    :: A(d,d), B(d,d), C(d**2,d**2)

  DO i=1,d
    DO k=1,d
      row=d*(i-1)+k
      DO j=1,d
        DO l=1,d
          col=d*(j-1)+l
          C(row,col)=A(i,j)*B(k,l)
        END DO
      END DO
    END DO
  END DO

  RETURN
END SUBROUTINE KRON

!***********************************************************************

SUBROUTINE Causal_Check
! Checks for causality of obtained solution. If noncausal, program will 
! terminate without likelihood computations.

  INTEGER                       :: i, j, k, t, km
  INTEGER                       :: lag(m)
  TYPE (matrix)                 :: phi(m)
  DOUBLE PRECISION, ALLOCATABLE :: A(:,:)  
  COMPLEX (KIND=8), ALLOCATABLE :: eig(:)
! kind=8 above specifies that eig's entries be double precision (LIB only).
  LOGICAL                       :: causal

  lag=orig_lags
  phi=topA(1:m)
  km=lag(m)
  ALLOCATE (A(2*km,2*km), eig(2*km))

! Build the matrix A ie. can write a VAR(p) as a VAR(1). 
  A=0
  DO i=1,2*km
    IF (i<3) THEN
      DO j=1,m
        A(i,2*lag(j)-1)=phi(j)%mat(i,1)
        A(i,2*lag(j))  =phi(j)%mat(i,2)
      END DO
    ELSE ! i=>3
      A(i,i-2)=1.0
    END IF
  END DO

! For the Phi's to be causal, simply check that all 2*km eigenvalues 
! of A are < 1 in absolute value. 
  CALL DEVLRG(2*km, A, 2*km, eig) 
  causal=.TRUE.
  DO i=1,2*km
    IF (ABS(eig(i)) >= 1.0) THEN !solution non-causal
      causal=.FALSE.
      EXIT
    END IF
  END DO

  IF (causal) THEN !if solution causal, return control to main prog.
    DEALLOCATE (A, eig)
    RETURN 
  ELSE                        !solution non-causal, prog. will terminate.
    PRINT*,"%%% NON-CAUSAL SOLUTION. PROGRAM WILL TERMINATE. %%%"
    PRINT*,"****************************************************"
    DEALLOCATE (A, eig)   
    STOP                
  END IF

END SUBROUTINE Causal_Check

!***********************************************************************

SUBROUTINE likelihood(what) 
! Exact likelihood calculation: very, very slow...
! First computes ACVF of a 2d subset AR model with m coeffts (phi) and lags 
! (lag), sigma^2=s2, into G(1-2km),...,G(-1),G(0),G(1),...,G(km-1).
! Then it gets -2 log  Likelihood for the vector of obs x: eqtn (11.5.5)
! in Yellow book, using the multivariate innnovations algorithm.

  INTEGER          :: i, j, k, t, km, drow
  INTEGER          :: lag(m)
  TYPE (vector)    :: xh(n)
  TYPE (matrix)    :: phi(m), s2, V(0:n-1), Vi(0:n-1)
  TYPE (matrix)    :: Ka(n,n), Th(n-1,n-1)
  DOUBLE PRECISION :: Z2(2,2), Tp(2,2), SumLogDetV, Term3, Like, temph(2)
  DOUBLE PRECISION :: tempv(2), DetV
  TYPE (matrix), ALLOCATABLE    :: G(:)
  DOUBLE PRECISION, ALLOCATABLE :: A(:,:), B(:,:), C(:,:), D(:,:) 
  CHARACTER        :: what*7

  lag=orig_lags
  phi=topA(1:m)
  s2=topvf
  km=lag(m)
  pi=3.141592654
  Z2=0.0
  ALLOCATE (G(1-2*km:km-1),A(2*km,2*km),B(4*km**2,4*km**2), &
            C(4*km**2,4*km**2),D(4*km**2,4*km**2))

! Begin --- building the covariances Gamma(h)
! Build the matrix A 
  A=0
  DO i=1,2*km
    IF (i<3) THEN
      DO j=1,m
        A(i,2*lag(j)-1)=phi(j)%mat(i,1)
        A(i,2*lag(j))  =phi(j)%mat(i,2)
      END DO
    ELSE ! i=>3
      A(i,i-2)=1.0
    END IF
  END DO
! B=A kronecker A
  CALL KRON(2*km,A,A,B)
! Form C = I - B
  C=-B
  DO i=1,4*km**2
    C(i,i)=1.0-B(i,i)
  END DO
! D = inv(C)
  CALL DLINRG(4*km**2, C, 4*km**2, D, 4*km**2)
! vec G_y(0) = D vec(s2). This forms G(1-km),...,G(-1),G(0),G(1),...,G(km-1)
  DO k=0,km-1
   DO i=1,2
    DO j=1,2
      drow=2*k+i+2*(j-1)*km
      G(-k)%mat(i,j)=D(drow,1)*s2%mat(1,1)+D(drow,2*km+2)*s2%mat(2,2) &
                    +(D(drow,2)+D(drow,2*km+1))*s2%mat(1,2)
    END DO
   END DO
   G(k)%mat=TRANSPOSE(G(-k)%mat)
  END DO
! Now compute G(-km),...,G(-(2km-1))  
  DO k=km,2*km-1
    G(-k)%mat=0
    DO j=1,m
      G(-k)%mat=G(-k)%mat+MATMUL(G(lag(j)-k)%mat,TRANSPOSE(phi(j)%mat))
    END DO
  END DO
! print
!  DO k=1-2*km,2*km-1
!   PRINT*,"Gamma(",k,"):"
!   CALL DWRRRL('',2,2, G(k)%mat,2, 0,'(W20.6)','NONE','NONE')
!  END DO 

! Begin --- building the K(i,j)'s, recursions (11.4.27)
  DO i=1,n
    DO j=1,n
      Ka(i,j)%mat=Z2
    END DO 
  END DO  
  DO i=1,n
    DO j=i,n
      IF (j<=km) Ka(i,j)%mat=G(i-j)%mat
      IF (i<=km .AND. km<j .AND. j<=2*km) THEN
        Ka(i,j)%mat=G(i-j)%mat
        DO k=1,m
          Ka(i,j)%mat=Ka(i,j)%mat-MATMUL(G(i-j+lag(k))%mat, &
                                         TRANSPOSE(phi(k)%mat))
        END DO      
      END IF
      IF (i==j .AND. i>km) Ka(i,j)%mat=s2%mat
    END DO
  END DO
  DO j=1,n
    DO i=j+1,n
      Ka(i,j)%mat=TRANSPOSE(Ka(j,i)%mat)
    END DO      
  END DO

! Begin --- building the theta(i,j)'s & V's, recursions (11.4.23)
  V(0)%mat=Ka(1,1)%mat  
! Store inverses of V, the Vi's
  CALL DLINRG(2, V(0)%mat, 2, Vi(0)%mat, 2)
! Initialize & Keep total of sum of log(det(V))'s
  DetV=V(0)%mat(1,1)*V(0)%mat(2,2)-V(0)%mat(2,1)*V(0)%mat(1,2)
  SumLogDetV=log(DetV)
  DO t=1,n-1
    DO k=0,t-1
      Tp=0
      DO j=0,k-1
        Tp=Tp+MATMUL(MATMUL(Th(t,t-j)%mat,V(j)%mat),TRANSPOSE(Th(k,k-j)%mat))
      END DO        
      Th(t,t-k)%mat=MATMUL((Ka(t+1,k+1)%mat-Tp),Vi(k)%mat)
    END DO 
    Tp=0
    DO j=0,t-1
      Tp=Tp+MATMUL(MATMUL(Th(t,t-j)%mat,V(j)%mat),TRANSPOSE(Th(t,t-j)%mat))
    END DO  
!   Get V's, their inverses Vi's, and keep total of sum of log((det(V))'s
    V(t)%mat=Ka(t+1,t+1)%mat - Tp
    CALL DLINRG(2, V(t)%mat, 2, Vi(t)%mat, 2)
    DetV=V(t)%mat(1,1)*V(t)%mat(2,2)-V(t)%mat(2,1)*V(t)%mat(1,2)
    SumLogDetV=SumLogDetV+log(DetV)
  END DO

! Get the one step predictors xh's, (11.4.28)
  xh(1)%vec=0
  Term3=DOT_PRODUCT(x(1)%vec,MATMUL(Vi(0)%mat,x(1)%vec))
  DO t=1,n-1
    xh(t+1)%vec=0
    IF (t<km) THEN
     DO j=1,t
      xh(t+1)%vec=xh(t+1)%vec+MATMUL(Th(t,j)%mat,x(t+1-j)%vec-xh(t+1-j)%vec)
     END DO
    ELSE ! t=>km
     DO j=1,m
      xh(t+1)%vec=xh(t+1)%vec+MATMUL(phi(j)%mat,x(t+1-lag(j))%vec)
     END DO
    END IF
!   Keep total of Term3 = sum_{t=1}^n [(x(t)-xh(t))'Vi(t-1)(x(t)-xh(t))]
    Term3=Term3+DOT_PRODUCT(x(t+1)%vec-xh(t+1)%vec,  &
                            MATMUL(Vi(t)%mat,x(t+1)%vec-xh(t+1)%vec))
  END DO

! Finally: -2 log Likelihood = Like
  Like=2.0*n*log(2.0*pi)+SumLogDetV+Term3

  DO t=0,km-1
    PRINT*,"Gamma(",t,"):"
    CALL DWRRRL('',2,2,G(t)%mat,2,0,'(F20.6)','NONE','NONE') 
  END DO  

  PRINT*,"-2 Log Like (",what,") : ", Like
!  PRINT*,"AICC (Burg WN):        ", Like+4.0*n*(4.0*km+1)/(2.0*(n-1)-4.0*km)

  DEALLOCATE (G,A,B,C,D)

  RETURN
END SUBROUTINE likelihood

!***********************************************************************

SUBROUTINE approx_likelihoods(truevf) 
! Computes approx -2 log likelihood for the algorithm obtained
! Phi's (topA), and WN variance (topvf). 
! Also computes WN variance estimate (RSS/n) starting with topvf WN estimate,
! that minimizes the -2 log likelihood for the algorithm obtained Phi's.

  INTEGER          :: i, maxitn
  DOUBLE PRECISION :: ff, tzz(3), step, oldff, truevf(3)
  DOUBLE PRECISION :: like1, like2, like

! Get likelihood for the WN variance estimate
  tzz(1)=topvf%mat(1,1); tzz(2)=topvf%mat(2,2); tzz(3)=topvf%mat(1,2)  
  CALL lad_fun(3, tzz, oldff)
!  PRINT*,"AICC (Burg WN):        ", ff+4.0*n*(4.0*km+1)/(2.0*(n-1)-4.0*km)
  PRINT*,"-2 Log Like (",stamp,"): ", oldff

! Now Get WN estimate that maximizes likelihood, for the algorithm Phi's.
! This is equivalent to RSS/n in one dimension. First start search with
! topvf estimate from algo
  step=0.1   
  CALL Hooke(3, tzz, step, ff)
  DO WHILE (ABS(ff-oldff) > 0.0001)
    oldff=ff
    step=step/10.0
    CALL Hooke(3, tzz, step, ff)
  END DO
  like1=MIN(ff, oldff)

! then use truevf (if it exists)
  IF (truevf(1)>0) THEN
    CALL lad_fun(3, truevf, oldff)
    step=0.1   
    CALL Hooke(3, truevf, step, ff)
    DO WHILE (ABS(ff-oldff) > 0.0001)
      oldff=ff
      step=step/10.0
      CALL Hooke(3, truevf, step, ff)
    END DO
    like2=MIN(ff, oldff)
  ELSE
    like2=1.0D30
  END IF

! Now take lowest like as the RSS/n 
  IF (like1 > like2) THEN
    like=like2
    tzz=truevf
  ELSE ! like1 smaller
    like=like1
  END IF 
  PRINT*,"-2 Log Like (RSS/n): ", like

! get the RSS/n matrix
  topvf%mat(1,1)=tzz(1); topvf%mat(2,2)=tzz(2)
  topvf%mat(1,2)=tzz(3); topvf%mat(2,1)=tzz(3)
  PRINT*," "
  PRINT*,"Estimated RSS/n (forward) WN covariance matrix:"
  CALL DWRRRL('',2,2,topvf%mat,2,0,'(F20.6)','NONE','NONE')
  PRINT*,"%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%"

  RETURN
END SUBROUTINE approx_likelihoods

!***********************************************************************

SUBROUTINE lad_fun(nn, zz, ff) 
! *** Objective function to go with H&J routine  ***
! Computes approx -2 log likelihood for the given Phi's (phi), and the 3 
! components of the WN variance (zz). Uses recursions 11.3.12 (YB) to find
! Psi's. G(h) found by truncated summation h=0,...,l (instead of infty).

  INTEGER          :: i, j, k, t, km, l, h, nn 
  INTEGER          :: lag(m)
  TYPE (matrix)    :: phi(m), s2, s2i
  DOUBLE PRECISION :: t1, t2, t3, t4, tvec(2), zz(*), ff
  DOUBLE PRECISION, ALLOCATABLE    :: GK(:,:), GKi(:,:), y(:)
  TYPE (matrix),    ALLOCATABLE    :: G(:), Psi(:), Phy(:)

  l=100          ! the truncation for the ACVF's from the PSi's
  lag=orig_lags
  phi=topA(1:m)
  km=lag(m)
  pi=3.141592654

! Build in constraints for +ve def. WN:
  IF ((zz(1)<0).OR.(zz(2)<0).OR.(zz(1)*zz(2)<zz(3)**2)) THEN
    ff=1.0D30
    RETURN
  END IF
 
!  IF (zz(1)<0.0) zz(1)=ABS(zz(1))
!  IF (zz(2)<0.0) zz(2)=ABS(zz(2))
!  IF (zz(1)*zz(2)<zz(3)**2) zz(3)=SQRT(zz(1)*zz(2))/2.0

! Build the WN, s2
  s2%mat(1,1)=zz(1); s2%mat(2,2)=zz(2)
  s2%mat(2,1)=zz(3); s2%mat(1,2)=zz(3)

  ALLOCATE (G(0:km-1), GK(2*km,2*km), Psi(0:km-1+l), Phy(1:km))
  ALLOCATE (y(2*km), GKi(2*km,2*km)) 

! Need to form Phy=0 except at lag(1:m)
  DO h=1,km
    Phy(h)%mat(1,:)=(/0.0,0.0/); Phy(h)%mat(2,:)=(/0.0,0.0/)
  END DO  
  DO t=1,m
    Phy(lag(t))%mat=phi(t)%mat
  END DO  

! Now get the Psi's
  Psi(0)%mat(1,:)=(/1.0,0.0/); Psi(0)%mat(2,:)=(/0.0,1.0/)
  DO j=1,km-1+l
    Psi(j)%mat(1,:)=(/0.0,0.0/); Psi(j)%mat(2,:)=(/0.0,0.0/)
    DO t=1,MIN(j,km)
      Psi(j)%mat=Psi(j)%mat+MATMUL(Phy(t)%mat,Psi(j-t)%mat)    
    END DO
  END DO

! Now compute covariance matrices
  DO h=0,km-1
    G(h)%mat(1,:)=(/0.0,0.0/); G(h)%mat(2,:)=(/0.0,0.0/)
    DO j=0,l
      G(h)%mat=G(h)%mat+MATMUL(MATMUL(Psi(h+j)%mat,s2%mat), &
                                             TRANSPOSE(Psi(j)%mat))
    END DO
!    PRINT*,"Truncated Gamma(",h,"), with l=",l,":"
!    CALL DWRRRL('',2,2,G(h)%mat,2,0,'(F20.6)','NONE','NONE') 
  END DO

! Form Big covariance matrix GK (symmetric):
  DO i=1,km
    DO j=1,i
      DO h=0,1
        DO k=0,1
          GK(2*i-h,2*j-k)=G(i-j)%mat(2-h,2-k)
          GK(2*j-k,2*i-h)=GK(2*i-h,2*j-k)
        END DO
      END DO
    END DO
  END DO

! Form term4
  t4=0.0
  CALL DLINDS(2, s2%mat, 2, s2i%mat, 2)  
  DO t=1+km,n
    tvec=0.0
    DO k=1,m
      tvec=tvec+MATMUL(phi(k)%mat,x(t-lag(k))%vec)
    END DO
    tvec=x(t)%vec-tvec
    t4=t4+DOT_PRODUCT(tvec,MATMUL(s2i%mat,tvec))    
  END DO

! Form t3
  CALL DLINDS(2*km, GK, 2*km, GKi, 2*km)
  DO j=1,km
    DO k=0,1
      y(2*j-k)=x(j)%vec(2-k)
    END DO
  END DO
  t3=DOT_PRODUCT(y,MATMUL(GKi,y))

! Form t2
  t2=(n-km)*LOG(s2%mat(1,1)*s2%mat(2,2)-s2%mat(1,2)**2)

! Form t1
! find det of GK by Choleski decomposing GK->GKi (re-use to save space), 
! then det(GK)=(product of diagonal elements of GKi)**2.
  CALL DLFTDS(2*km,GK,2*km,GKi,2*km)
  t1=0.0
  DO h=1,2*km
    t1=t1+LOG(GKi(h,h))
  END DO
  t1=2*t1
  
! Now put it all together
  ff=2.0*n*log(2.0*pi)+t1+t2+t3+t4

  DEALLOCATE (G, GK, GKi, Psi, Phy, y)

  RETURN
END SUBROUTINE lad_fun

!*********************************************************************
!subroutine optimization using Hooke and Jeeves
SUBROUTINE Hooke(p,AR,er,fv)

	INTEGER p
	DOUBLE PRECISION AR(p),er,fv
	
	INTEGER flg,fi,ik,m
	DOUBLE PRECISION x1(p),x2(p),c(p),bx(p)
	DOUBLE PRECISION ac, min,bmin
	intent(inout) :: AR,er

		m=p
		ac=.4*er
		fi=0
		ik=0
		ij=0
		iq=1
		x1=AR

 3073   do 3075 i=1,m
 3075   c(i)=x1(i)
! 3080   if(iq.eq.0)then
!        write(*,*) '      <Computing Gaussian likelihood>'
!        endif
        call lad_fun(p,c,fv)
        min=fv
 3085   if(iq.eq.0)goto 3640
        goto 3600
!
!       Exploratory moves
!
 3100   do 3105 j=1,m
 3105   x2(j)=x1(j)
        do 3160 i=1,m
        do 3115 j=1,m
 3115   c(j)=x2(j)
        c(i)=x1(i)+er
        call lad_fun(p,c,fv)
 3120   if(fv.lt.min) then
           min=fv
           x2(i)=c(i)
        endif
        c(i)=x1(i)-er
        call lad_fun(p,c,fv)
 3130   if(fv.lt.min) then
           min=fv
           x2(i)=c(i)
        endif
 3160   continue
        flg=1
        do 3180 i=1,m
        if(x1(i).ne.x2(i))flg=0
 3180   continue
        if(flg.eq.0)goto 3195
		er=er/2
        if(er.lt.ac)fi=1
        if(ik.eq.0)goto 3285
        if(ik.eq.1)goto 3605
        goto 3100
 3195   if(ik.eq.0)goto 3285
        if(ik.eq.1)goto 3605
!
!       Pattern moves
!
 3200   do 3230 i=1,m
 3230   x1(i)=2*x2(i)-x1(i)
        do 3260 i=1,m
 3260   bx(i)=x2(i)
        !bvar=var
        bmin=min
        do 3276 i=1,m
 3276   c(i)=x1(i)
        call lad_fun(p,c,fv)
 3280   min=fv
        ik=0
        goto 3100
 3285   if(fi.eq.1)goto 3615
        if(min.ge.bmin)goto 3310
        goto 3200
 3310   do 3330 i=1,m
 3330   x1(i)=bx(i)
        !var=bvar
        min=bmin
        goto 3615
!
!       End of moves
!
 3600   ik=1
        goto 3100
 3605   if(fi.eq.1)goto 3640
 3610   goto 3200
 3615   if(fi.eq.1)goto 3640
        goto 3600
 3640   continue
!3680   fi=0
! 3687   d=(.5/sqrt(xn))/(5**iii)
 3647    continue
        if(iq.ne.0)then
         min=bmin
 4310    do 4330 i=1,p
           IF (i .le. p) AR(i)=bx(i)
 4330    x1(i)=bx(i)
         fv=min
        endif	
	
	END SUBROUTINE Hooke

!***********************************************************************



END MODULE tree

!***********************************************************************
!***********************************************************************

PROGRAM Bdt2

! this program subset models AR(p)'s for 2-dimensional time series, using 
! Burg type recursions, with subset size <=26, km<100, and a max of 10000 obs.

  USE tree
  INTEGER                                     :: i, wn_known, mc
  INTEGER, ALLOCATABLE, DIMENSION (:)         :: toplags
  TYPE (vector)                               :: y(10000)
  CHARACTER                                   :: h*24
  DOUBLE PRECISION                            :: mean1, mean2, truevf(3)

  PRINT*, "%%%%%%%%%%%%%%% Bivariate SVAR Modeling Program %%%%%%%%%%%%%%%"
! read in the time series from a data file
  20   write(*,*)
       write(*,22)
  22    format(5x,'Enter file name of time series for modelling: ',$)
  23    h='                        '
        read(*,*) h
        IF (h=='a                       ') h='sun2.tsm'
        open(3,file=h,status='old',err=20)
        i=1
  25    read(3,*,end=30) y(i)%vec(1), y(i)%vec(2)
        i=i+1
          if (i.eq.10002) then
           write(*,6665)
6665       format(3x,'DATA TRUNCATED AFTER FIRST 10000 OBSERVATIONS.')
           i=i-1
           goto 30
          endif
        goto 25
  30    n=i-1
        close(3)

! now mean-correct the obs
  mean1=SUM(y%vec(1))/n
  mean2=SUM(y%vec(2))/n
  ALLOCATE (x(n))  
  x%vec(1)=(/(y(i)%vec(1), i=1,n)/)
  x%vec(2)=(/(y(i)%vec(2), i=1,n)/)
  PRINT*, "Do you wish to mean-correct the observations (1=yes, 0=no)?"
  READ*, mc
  IF (mc==1) THEN 
    x%vec(1)=(/(y(i)%vec(1)-mean1, i=1,n)/)
    x%vec(2)=(/(y(i)%vec(2)-mean2, i=1,n)/)    
  END IF
  PRINT*, "There are ",n," observations."
  WRITE(*,40) x(1), x(n)
  40   FORMAT("Firs obs is ",F20.4," last is ",F20.4)

! Enter how many lags will be modeled
  PRINT*, "Enter the number of lags to be modeled (<27):"
  READ*, m
  ALLOCATE (toplags(m))

! read in the lags
  PRINT*, "Enter the lags:"
  READ*, (toplags(i), i=1,m)
! PRINT*,"The lags are: ", (toplags(i), i=1,m)

! Simulation?
  PRINT*, "Is the true white noise covariance matrix known (1=yes, 0=no)?"
  READ*, wn_known
  IF (wn_known==1) THEN
    PRINT*, "Enter Sigma(1,1), Sigma(2,2), Sigma(1,2):"
    READ*, (truevf(i), i=1,3)    
  ELSE 
! don't know true WN, so set truevf(1)=0 as flag for routine approx_likelihoods
    truevf(1)=0    
  END IF

! method
  PRINT*, "Enter the method for obtaining the reflection coefficients."
  PRINT*, "Yule-Walker (1), Burg (2), Vieira-Morf (3), Nuttal-Strand (4):"
  READ*, method  
  SELECT CASE (method)
    CASE (1); stamp="YuWa"
    CASE (2); stamp="Burg"
    CASE (3); stamp="Morf"
    CASE (4); stamp="Nutt"
    CASE DEFAULT
      PRINT*,"Method not in range: ", method   
      STOP
  END SELECT 
  
  CALL make_tree(toplags, truevf)

END PROGRAM Bdt2      


